"""Composer models
"""
from django.db import models
from core_main_app.permissions.utils import get_formatted_name
from core_composer_app.permissions import rights


class Composer(models.Model):
    class Meta:
        verbose_name = 'core_composer_app'
        default_permissions = ()
        permissions = (
            (rights.composer_access, get_formatted_name(rights.composer_access)),
            (rights.composer_save_template, get_formatted_name(rights.composer_save_template)),
            (rights.composer_save_type, get_formatted_name(rights.composer_save_type)),
        )



"""Composer admin views
"""
from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.core.urlresolvers import reverse
from django.http.response import HttpResponseRedirect
from django.shortcuts import redirect
from django.template import loader
from django.utils.html import escape as html_escape

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket
from core_composer_app.components.type.models import Type
from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_composer_app.views.admin.ajax import EditBucketView
from core_composer_app.views.admin.forms import BucketForm, UploadTypeForm, EditTypeBucketsForm
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import NotUniqueError
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.utils.rendering import admin_render
from core_main_app.utils.xml import get_imports_and_includes
from core_main_app.views.admin.forms import UploadVersionForm
from core_main_app.views.common.ajax import EditTemplateVersionManagerView
from core_main_app.views.common.views import read_xsd_file
from core_main_app.views.user.views import get_context_manage_template_versions


@staff_member_required
def manage_types(request):
    """View that allows type management.

    Args:
        request:

    Returns:

    """
    # get all types
    type_version_managers = type_version_manager_api.get_global_version_managers()
    # get buckets
    buckets = bucket_api.get_all()

    context = {
        'object_name': "Type",
        'available': [type_version_manager for type_version_manager in type_version_managers
                      if not type_version_manager.is_disabled],
        'disabled': [type_version_manager for type_version_manager in type_version_managers
                     if type_version_manager.is_disabled],
        'buckets': buckets
    }

    assets = {
        "js": [
            {
                "path": 'core_main_app/common/js/templates/list/restore.js',
                "is_raw": False
            },
            {
                "path": 'core_main_app/common/js/templates/list/modals/disable.js',
                "is_raw": False
            },
            EditTemplateVersionManagerView.get_modal_js_path()
        ],
        "css": ['core_composer_app/common/css/bucket.css']
    }

    modals = [
        "core_main_app/admin/templates/list/modals/disable.html",
        EditTemplateVersionManagerView.get_modal_html_path()
    ]

    return admin_render(request,
                        'core_composer_app/admin/types/list.html',
                        assets=assets,
                        context=context,
                        modals=modals)


@staff_member_required
def manage_type_versions(request, version_manager_id):
    """View that allows type versions management.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    try:
        # get the version manager
        version_manager = version_manager_api.get(version_manager_id)
        context = get_context_manage_template_versions(version_manager)

        # updating context regarding the installed apps
        # default back_url initialization
        context.update({"back_url": "admin:core_composer_app_types"})
        if 'core_parser_app' in settings.INSTALLED_APPS:
            context.update({"module_url": "admin:core_composer_app_type_modules"})
        if 'core_dashboard_app' in settings.INSTALLED_APPS:
            # the dashboard exposes the user's version managers
            # in this view, we come from the dashboard
            if version_manager.user:
                context.update({"back_url": "admin:core_dashboard_types"})

        assets = {
            "js": [
                {
                    "path": 'core_main_app/common/js/templates/versions/set_current.js',
                    "is_raw": False
                },
                {
                    "path": 'core_main_app/common/js/templates/versions/restore.js',
                    "is_raw": False
                },
                {
                    "path": 'core_main_app/common/js/templates/versions/modals/disable.js',
                    "is_raw": False
                }
            ]
        }

        modals = ["core_main_app/admin/templates/versions/modals/disable.html"]

        return admin_render(request,
                            'core_composer_app/admin/types/versions.html',
                            assets=assets,
                            modals=modals,
                            context=context)
    except Exception, e:
        return admin_render(request,
                            'core_main_app/common/commons/error.html',
                            context={'error': e.message})


@staff_member_required
def upload_type(request):
    """Upload type.

    Args:
        request:

    Returns:

    """
    assets = {
        "js": [
            {
                "path": 'core_main_app/admin/js/templates/upload/dependencies.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/types/upload/dependencies.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/types/upload/dependency_resolver.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/types/upload/dependencies.raw.js',
                "is_raw": True
            }
        ]
    }

    context = {
        'object_name': "Type",
        'url': reverse("admin:core_composer_app_upload_type"),
        'redirect_url': reverse("admin:core_composer_app_types")
    }

    # method is POST
    if request.method == 'POST':
        form = UploadTypeForm(request.POST, request.FILES)
        context['upload_form'] = form

        if form.is_valid():
            return _save_type(request, assets, context)
        else:
            # Display error from the form
            return _upload_type_response(request, assets, context)
    # method is GET
    else:
        # render the form to upload a template
        context['upload_form'] = UploadTypeForm()
        return _upload_type_response(request, assets, context)


def _save_type(request, assets, context):
    """Save a type.

    Args:
        request:
        assets:
        context:

    Returns:

    """
    # get the schema name
    name = request.POST['name']
    # get the file from the form
    xsd_file = request.FILES['xsd_file']
    # read the content of the file
    xsd_data = read_xsd_file(xsd_file)
    # get the buckets
    buckets = request.POST.getlist('buckets')

    try:
        type_object = Type(filename=xsd_file.name, content=xsd_data)
        type_version_manager = TypeVersionManager(title=name)
        type_version_manager_api.insert(type_version_manager, type_object, buckets)
        return HttpResponseRedirect(reverse("admin:core_composer_app_types"))
    except exceptions.XSDError, xsd_error:
        return _handle_xsd_errors(request, assets, context, xsd_error, xsd_data, xsd_file.name)
    except Exception, e:
        context['errors'] = html_escape(e.message)
        return _upload_type_response(request, assets, context)


def _handle_xsd_errors(request, assets, context, xsd_error, xsd_content, filename):
    """Handle XSD errors. Builds dependency resolver if needed.

    Args:
        request:
        context:
        xsd_error:
        xsd_content:
        filename:

    Returns:

    """
    imports, includes = get_imports_and_includes(xsd_content)
    # a problem with includes/imports has been detected
    if len(includes) > 0 or len(imports) > 0:
        # build dependency resolver
        context['dependency_resolver'] = _get_dependency_resolver_html(imports, includes, xsd_content,
                                                                       filename)
        return _upload_type_response(request, assets, context)
    else:
        context['errors'] = html_escape(xsd_error.message)
        return _upload_type_response(request, assets, context)


def _get_dependency_resolver_html(imports, includes, xsd_data, filename):
    """Return HTML for dependency resolver form.

    Args:
        imports:
        includes:
        xsd_data:
        filename:

    Returns:

    """
    # build the list of dependencies
    current_types = type_version_manager_api.get_global_version_managers()
    list_dependencies_template = loader.get_template('core_main_app/admin/list_dependencies.html')
    context = {'templates': [template for template in current_types if not template.is_disabled]}
    list_dependencies_html = list_dependencies_template.render(context)

    # build the dependency resolver form
    dependency_resolver_template = loader.get_template('core_main_app/admin/dependency_resolver.html')
    context = {
        'imports': imports,
        'includes': includes,
        'xsd_content': html_escape(xsd_data),
        'filename': filename,
        'dependencies': list_dependencies_html,
    }
    return dependency_resolver_template.render(context)


def _upload_type_response(request, assets, context):
    """Render type upload response.

    Args:
        request:
        context:

    Returns:

    """
    return admin_render(request,
                        'core_composer_app/admin/types/upload.html',
                        assets=assets,
                        context=context)


@staff_member_required
def upload_type_version(request, version_manager_id):
    """Upload type version.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    assets = {
        "js": [
            {
                "path": 'core_main_app/admin/js/templates/upload/dependencies.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/types/upload/dependencies.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/types/upload/dependency_resolver.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/types/upload/dependencies.raw.js',
                "is_raw": True
            }
        ]
    }

    type_version_manager = version_manager_api.get(version_manager_id)
    context = {
        'object_name': "Type",
        'version_manager': type_version_manager,
        'url': reverse("admin:core_composer_app_upload_type_version",
                       kwargs={'version_manager_id': type_version_manager.id}),
        'redirect_url': reverse("admin:core_composer_app_manage_type_versions",
                                kwargs={'version_manager_id': type_version_manager.id})
    }

    # method is POST
    if request.method == 'POST':
        form = UploadVersionForm(request.POST, request.FILES)
        context['upload_form'] = form

        if form.is_valid():
            return _save_type_version(request, assets, context, type_version_manager)
        else:
            # Display errors from the form
            return _upload_type_response(request, assets, context)
    # method is GET
    else:
        # render the form to upload a template
        context['upload_form'] = UploadVersionForm()
        return _upload_type_response(request, assets, context)


def _save_type_version(request, assets, context, type_version_manager):
    """Save a type version.

    Args:
        request:
        assets:
        context:
        type_version_manager:

    Returns:

    """
    # get the file from the form
    xsd_file = request.FILES['xsd_file']
    # read the content of the file
    xsd_data = read_xsd_file(xsd_file)

    try:
        type_object = Type(filename=xsd_file.name, content=xsd_data)
        type_version_manager_api.insert(type_version_manager, type_object)
        return HttpResponseRedirect(reverse("admin:core_composer_app_manage_type_versions",
                                            kwargs={'version_manager_id': str(type_version_manager.id)}))
    except exceptions.XSDError, xsd_error:
        return _handle_xsd_errors(request, assets, context, xsd_error, xsd_data, xsd_file.name)
    except Exception, e:
        context['errors'] = html_escape(e.message)
        return _upload_type_response(request, assets, context)


@staff_member_required
def manage_buckets(request):
    """Manage buckets view.

    Args:
        request:

    Returns:

    """

    context = {
        'object_name': "Bucket",
        'buckets': bucket_api.get_all()
    }

    assets = {
        "js": [
            {
                "path": 'core_composer_app/admin/js/bucket.js',
                "is_raw": False
            },
            {
                "path": 'core_composer_app/admin/js/bucket.raw.js',
                "is_raw": True
            },
            {
                "path": 'core_main_app/common/js/backtoprevious.js',
                "is_raw": True
            },
            EditBucketView.get_modal_js_path(),
        ],
        "css": ['core_composer_app/common/css/bucket.css']
    }

    modals = ['core_composer_app/admin/buckets/modals/delete.html',
              EditBucketView.get_modal_html_path()]

    return admin_render(request,
                        'core_composer_app/admin/buckets/list.html',
                        assets=assets,
                        modals=modals,
                        context=context)


@staff_member_required
def upload_bucket(request):
    """Upload bucket.

    Args:
        request:

    Returns:

    """
    assets = {
        "js": [
            {
                "path": 'core_main_app/common/js/backtoprevious.js',
                "is_raw": True
            }
        ]
    }

    context = {
        'object_name': 'Bucket'
    }
    if request.method == 'POST':
        form = BucketForm(request.POST)
        if form.is_valid():
            bucket_label = request.POST['label']
            bucket = Bucket(label=bucket_label)
            try:
                bucket_api.upsert(bucket)
                return redirect(reverse('admin:core_composer_app_buckets'))
            except NotUniqueError:
                context['errors'] = "A bucket with the same name already exists."
            except Exception, e:
                context['errors'] = e.message

    else:
        form = BucketForm()

    context['form'] = form
    return admin_render(request,
                        'core_composer_app/admin/buckets/add.html',
                        context=context, assets=assets)


@staff_member_required
def manage_type_buckets(request, version_manager_id):
    """Manage buckets of a type.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    # get the version manager
    version_manager = None

    try:
        version_manager = version_manager_api.get(version_manager_id)
    except:
        # TODO: catch good exception, redirect to error page
        pass

    context = {
        'version_manager': version_manager,
        'buckets': bucket_api.get_all()
    }

    assets = {
        "css": ['core_composer_app/common/css/bucket.css'],
        "js": [
            {
                "path": 'core_main_app/common/js/backtoprevious.js',
                "is_raw": True
            }
        ]
    }
    if request.method == 'POST':
        form = EditTypeBucketsForm(request.POST)
        if form.is_valid():
            buckets = request.POST.getlist('buckets')
            bucket_api.update_type_buckets(version_manager, buckets)
            return redirect(reverse('admin:core_composer_app_types'))
    else:
        form = EditTypeBucketsForm()

    context['form'] = form
    return admin_render(request,
                        'core_composer_app/admin/buckets/edit.html',
                        assets=assets,
                        context=context)



"""REST views for the type version manager API
"""
from django.utils.decorators import method_decorator

from core_composer_app.components.type_version_manager import api as type_version_manager_api
from core_composer_app.rest.type_version_manager.abstract_views import AbstractTypeList
from core_main_app.rest.template_version_manager.views import AbstractTemplateVersionManagerList
from core_main_app.utils.decorators import api_staff_member_required


class GlobalTypeVersionManagerList(AbstractTemplateVersionManagerList):
    """ List all global type version managers.
    """

    def get_template_version_managers(self):
        """ Get global ty[e version managers

        Returns:

        """
        return type_version_manager_api.get_global_version_managers()


class UserTypeVersionManagerList(AbstractTemplateVersionManagerList):
    """ List all user type version managers.
    """

    def get_template_version_managers(self):
        """ Get all user type version managers.

        Returns:

        """
        return type_version_manager_api.get_version_managers_by_user(user_id=self.request.user.id)


class UserTypeList(AbstractTypeList):
    """ Create a user type.
    """
    def post(self, request):
        """ Create a user type.

        Args:
            request:

        Returns:

        """
        return super(UserTypeList, self).post(request)

    def get_user(self):
        return str(self.request.user.id)


class GlobalTypeList(AbstractTypeList):
    """ Create a global type.
    """
    @method_decorator(api_staff_member_required())
    def post(self, request):
        """ Create a global type.

        Args:
            request:

        Returns:

        """
        return super(GlobalTypeList, self).post(request)

    def get_user(self):
        return None



"""
Type Version Manager model
"""
from core_main_app.components.template_version_manager.models import TemplateVersionManager
from core_main_app.components.version_manager.models import VersionManager


class TypeVersionManager(TemplateVersionManager):
    """Manage versions of types.
    """

    # TODO: see if better way to find _cls
    class_name = 'VersionManager.TemplateVersionManager.TypeVersionManager'

    @staticmethod
    def get_global_version_managers(_cls=True):
        """Return all Type Version Managers with user set to None.

        Returns:

        """
        return TypeVersionManager.objects(user=None).all()

    @staticmethod
    def get_active_global_version_manager(_cls=True):
        """Return all active Type Version Managers with user set to None.

        Returns:

        """
        return TypeVersionManager.objects(is_disabled=False, user=None).all()

    @staticmethod
    def get_version_managers_by_user(user_id):
        """Return Type Version Managers with user set to user_id.

        Args:
            user_id:

        Returns:

        """
        return TypeVersionManager.objects(user=str(user_id)).all()

    @staticmethod
    def get_active_version_manager_by_user_id(user_id):
        """Return active Type Version Managers with user set to user_id.

        Args:
            user_id:

        Returns:

        """
        return TypeVersionManager.objects(is_disabled=False, user=str(user_id)).all()

    @staticmethod
    def get_all_type_version_manager_except_user_id(user_id):
        """ Return all Version Managers of all users except user with given user id.

        Args:
            user_id:

        Returns:

        """
        return VersionManager.objects(_cls=TypeVersionManager.class_name, user__nin=str(user_id)).all()

    @staticmethod
    def get_all_type_version_manager():
        """ Return all Version Managers of all users.

        Args:

        Returns:

        """
        return VersionManager.objects(_cls=TypeVersionManager.class_name).all()



""" Integration Test for Bucket Rest API
"""

from rest_framework import status

from core_composer_app.rest.bucket import views
from core_main_app.utils.integration_tests.integration_base_test_case import \
    MongoIntegrationBaseTestCase
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import RequestMock
from tests.components.bucket.fixtures.fixtures import BucketFixtures

fixture_bucket = BucketFixtures()


class TestBucketList(MongoIntegrationBaseTestCase):
    fixture = fixture_bucket

    def setUp(self):
        super(TestBucketList, self).setUp()
        self.data = {'label': 'label'}

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(), user)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_all_buckets(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(), user)

        # Assert
        self.assertEqual(len(response.data), 3)

    def test_get_by_correct_label_returns_one_bucket(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(),
                                              user,
                                              data={'label': self.fixture.bucket_1.label})

        # Assert
        self.assertEqual(len(response.data), 1)

    def test_get_by_correct_label_returns_correct_bucket(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(),
                                              user,
                                              data={'label': self.fixture.bucket_1.label})

        # Assert
        self.assertEqual(response.data[0]['label'], self.fixture.bucket_1.label)

    def test_get_by_incorrect_label_returns_no_bucket(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketList.as_view(),
                                              user,
                                              data={'label': 'incorrect label'})

        # Assert
        self.assertEqual(len(response.data), 0)

    def test_post_returns_http_201(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_as_user_returns_http_403(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_post_empty_types_returns_http_201(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data['types'] = []

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_with_one_correct_type_returns_http_201(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data['types'] = [self.fixture.type_vm_1.id]

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_with_one_correct_type_returns_response_containing_one_type(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data['types'] = [self.fixture.type_vm_1.id]

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(len(response.data['types']), 1)

    def test_post_with_two_correct_type_returns_http_201(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data['types'] = [self.fixture.type_vm_1.id, self.fixture.type_vm_2.id]

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_post_with_one_correct_type_returns_response_containing_two_types(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data['types'] = [self.fixture.type_vm_1.id, self.fixture.type_vm_2.id]

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(len(response.data['types']), 2)

    def test_post_with_one_incorrect_type_returns_http_400(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data['types'] = ['wrong_type_id']

        # Act
        response = RequestMock.do_request_post(views.BucketList.as_view(),
                                               user,
                                               data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class TestBucketDetail(MongoIntegrationBaseTestCase):
    fixture = fixture_bucket

    def setUp(self):
        super(TestBucketDetail, self).setUp()
        self.data = {'label': 'label'}

    def test_get_returns_http_200(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketDetail.as_view(),
                                              user,
                                              param={'pk': str(self.fixture.bucket_1.id)})

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_returns_bucket(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketDetail.as_view(),
                                              user,
                                              param={'pk': str(self.fixture.bucket_1.id)})

        # Assert
        self.assertEqual(response.data['label'], self.fixture.bucket_1.label)

    def test_get_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user('1')

        # Act
        response = RequestMock.do_request_get(views.BucketDetail.as_view(),
                                              user,
                                              param={'pk': '507f1f77bcf86cd799439011'})

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_returns_http_204(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)

        # Act
        response = RequestMock.do_request_delete(views.BucketDetail.as_view(),
                                                 user,
                                                 param={'pk': str(self.fixture.bucket_1.id)})

        # Assert
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)

        # Act
        response = RequestMock.do_request_delete(views.BucketDetail.as_view(),
                                                 user,
                                                 param={'pk': '507f1f77bcf86cd799439011'})

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_returns_http_200(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)

        # Act
        response = RequestMock.do_request_patch(views.BucketDetail.as_view(),
                                                user,
                                                param={'pk': str(self.fixture.bucket_1.id)},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_returns_updated_label(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)

        # Act
        response = RequestMock.do_request_patch(views.BucketDetail.as_view(),
                                                user,
                                                param={'pk': str(self.fixture.bucket_1.id)},
                                                data=self.data)

        # Assert
        self.assertEqual(response.data['label'], self.data['label'])

    def test_patch_wrong_id_returns_http_404(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)

        # Act
        response = RequestMock.do_request_patch(views.BucketDetail.as_view(),
                                                user,
                                                param={'pk': '507f1f77bcf86cd799439011'})

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)


class TestTypeVersionManagerBuckets(MongoIntegrationBaseTestCase):
    fixture = fixture_bucket

    def setUp(self):
        super(TestTypeVersionManagerBuckets, self).setUp()

    def test_patch_returns_http_200(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data = [{'id': str(self.fixture.bucket_1.id)}]

        # Act
        response = RequestMock.do_request_patch(views.TypeVersionManagerBuckets.as_view(),
                                                user,
                                                param={'pk': self.fixture.type_vm_1.id},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_two_correct_bucket_id_returns_http_200(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data = [{'id': str(self.fixture.bucket_1.id)},
                     {'id': str(self.fixture.bucket_2.id)}]

        # Act
        response = RequestMock.do_request_patch(views.TypeVersionManagerBuckets.as_view(),
                                                user,
                                                param={'pk': self.fixture.type_vm_1.id},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_no_bucket_id_returns_http_200(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data = []

        # Act
        response = RequestMock.do_request_patch(views.TypeVersionManagerBuckets.as_view(),
                                                user,
                                                param={'pk': self.fixture.type_vm_1.id},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_wrong_bucket_id_returns_http_404(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data = [{'id': '507f1f77bcf86cd799439011'}]

        # Act
        response = RequestMock.do_request_patch(views.TypeVersionManagerBuckets.as_view(),
                                                user,
                                                param={'pk': self.fixture.type_vm_1.id},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_wrong_template_version_manager_id_returns_http_404(self):
        # Arrange
        user = create_mock_user('1', is_staff=True)
        self.data = [{'id': str(self.fixture.bucket_1.id)}]

        # Act
        response = RequestMock.do_request_patch(views.TypeVersionManagerBuckets.as_view(),
                                                user,
                                                param={'pk': '507f1f77bcf86cd799439011'},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_patch_other_user_type_version_manager_returns_http_403(self):
        # Arrange
        user = create_mock_user('2')
        self.data = [{'id': str(self.fixture.bucket_1.id)}]

        # Act
        response = RequestMock.do_request_patch(views.TypeVersionManagerBuckets.as_view(),
                                                user,
                                                param={'pk': self.fixture.type_vm_1.id},
                                                data=self.data)

        # Assert
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)



"""Serializers used throughout the bucket Rest API
"""
from django.http import Http404
from rest_framework_mongoengine.serializers import DocumentSerializer

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket
from rest_framework.fields import CharField

from core_main_app.commons.exceptions import DoesNotExist


class BucketSerializer(DocumentSerializer):
    """ Bucket serializer
    """
    class Meta:
        """ Meta
        """
        model = Bucket
        fields = '__all__'
        read_only_fields = ('id', 'color', )

    def create(self, validated_data):
        """
        Create and return a new `Bucket` instance, given the validated data.
        """
        # Create data
        bucket = Bucket(label=validated_data['label'],
                        types=validated_data.get('types', None))
        # Save the data
        return bucket_api.upsert(bucket)

    def update(self, instance, validated_data):
        """
        Update and return an existing `Bucket` instance, given the validated data.
        """
        instance.label = validated_data.get('label', instance.label)
        return bucket_api.upsert(instance)


class BucketsSerializer(DocumentSerializer):
    """ Buckets serializer.
    """
    id = CharField()

    class Meta:
        model = Bucket
        fields = ('id', )

    def validate_id(self, id):
        """ Validate id field

        Args:
            id:

        Returns:

        """
        try:
            bucket_api.get_by_id(id)
            return id
        except DoesNotExist:
            raise Http404



"""Bucket test cases
"""
from unittest.case import TestCase
from bson.objectid import ObjectId
from mock.mock import Mock, patch
from core_main_app.commons import exceptions

from core_composer_app.components.bucket.models import Bucket
from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager


class TestBucketGetById(TestCase):
    @patch.object(Bucket, 'get_by_id')
    def test_bucket_get_by_id_returns_bucket(self, mock_get_by_id):
        # Arrange
        mock_bucket = _create_mock_bucket()

        mock_get_by_id.return_value = mock_bucket

        # Act
        result = bucket_api.get_by_id(mock_bucket.id)

        # Assert
        self.assertIsInstance(result, Bucket)

    @patch.object(Bucket, 'get_by_id')
    def test_bucket_get_by_id_raises_exception_if_object_does_not_exist(self, mock_get_by_id):
        # Arrange
        mock_absent_id = ObjectId()
        mock_get_by_id.side_effect = exceptions.DoesNotExist('')

        # Act + Assert
        with self.assertRaises(exceptions.DoesNotExist):
            bucket_api.get_by_id(mock_absent_id)


class TestBucketUpsert(TestCase):
    @patch.object(Bucket, 'get_colors')
    @patch.object(Bucket, 'save')
    def test_upsert_bucket_returns_bucket(self, mock_save, mock_get_colors):
        bucket = _create_bucket()

        mock_save.return_value = bucket
        mock_get_colors.return_value = []
        result = bucket_api.upsert(bucket)
        self.assertIsInstance(result, Bucket)


class TestBucketGetAll(TestCase):
    @patch.object(Bucket, 'get_all')
    def test_get_all_buckets_returns_buckets(self, mock_get_all):
        # Arrange
        mock_bucket1 = _create_mock_bucket()
        mock_bucket2 = _create_mock_bucket()

        mock_get_all.return_value = [mock_bucket1, mock_bucket2]

        # Act
        result = bucket_api.get_all()

        # Assert
        self.assertTrue(all(isinstance(item, Bucket) for item in result))


class TestAddTypeToBuckets(TestCase):
    @patch.object(Bucket, 'get_by_id')
    @patch.object(Bucket, 'get_colors')
    @patch.object(Bucket, 'save')
    def test_add_type_to_buckets_adds_one_to_list(self, mock_save, mock_get_colors, mock_get_by_id):
        bucket = _create_bucket()
        mock_save.return_value = bucket
        mock_get_by_id.return_value = bucket
        mock_get_colors.return_value = []

        mock_version_manager = _create_mock_type_version_manager()

        self.assertTrue(len(bucket.types) == 0)
        bucket_api.add_type_to_buckets(mock_version_manager, [bucket.id])
        self.assertTrue(len(bucket.types) == 1)

    @patch.object(Bucket, 'get_by_id')
    @patch.object(Bucket, 'get_colors')
    @patch.object(Bucket, 'save')
    def test_add_no_type_to_buckets_does_not_update_bucket(self, mock_save, mock_get_colors, mock_get_by_id):
        bucket = _create_bucket()
        mock_save.return_value = bucket
        mock_get_by_id.return_value = bucket
        mock_get_colors.return_value = []

        mock_version_manager = _create_mock_type_version_manager()

        self.assertTrue(len(bucket.types) == 0)
        bucket_api.add_type_to_buckets(mock_version_manager, [])
        self.assertTrue(len(bucket.types) == 0)

    @patch.object(Bucket, 'get_by_id')
    @patch.object(Bucket, 'get_colors')
    @patch.object(Bucket, 'save')
    def test_add_type_to_buckets_raises_exception_if_bucket_id_not_found(self, mock_save, mock_get_colors,
                                                                         mock_get_by_id):

        mock_get_by_id.side_effect = exceptions.DoesNotExist
        mock_get_colors.return_value = []

        mock_version_manager = _create_mock_type_version_manager()

        with self.assertRaises(exceptions.ApiError):
            bucket_api.add_type_to_buckets(mock_version_manager, [ObjectId()])


class TestRemoveTypeFromBuckets(TestCase):
    @patch.object(Bucket, 'get_all')
    @patch.object(Bucket, 'get_by_id')
    @patch.object(Bucket, 'get_colors')
    @patch.object(Bucket, 'save')
    def test_remove_type_from_buckets_substracts_one_to_list(self, mock_save, mock_get_colors,
                                                             mock_get_by_id, mock_get_all):
        bucket = _create_bucket()
        mock_version_manager = _create_mock_type_version_manager()

        bucket.types = [mock_version_manager]
        mock_get_all.return_value = [bucket]

        mock_save.return_value = bucket
        mock_get_by_id.return_value = bucket
        mock_get_colors.return_value = []

        self.assertTrue(len(bucket.types) == 1)
        bucket_api.remove_type_from_buckets(mock_version_manager)
        self.assertTrue(len(bucket.types) == 0)

    @patch.object(Bucket, 'get_all')
    @patch.object(Bucket, 'get_by_id')
    @patch.object(Bucket, 'get_colors')
    @patch.object(Bucket, 'save')
    def test_removes_absent_type_from_buckets_does_not_update_bucket(self, mock_save, mock_get_colors,
                                                                     mock_get_by_id, mock_get_all):
        bucket = _create_bucket()
        mock_version_manager = _create_mock_type_version_manager()
        mock_absent_version_manager = _create_mock_type_version_manager()

        bucket.types = [mock_version_manager]
        mock_get_all.return_value = [bucket]

        mock_save.return_value = bucket
        mock_get_by_id.return_value = bucket
        mock_get_colors.return_value = []

        self.assertTrue(len(bucket.types) == 1)
        bucket_api.remove_type_from_buckets(mock_absent_version_manager)
        self.assertTrue(len(bucket.types) == 1)


def _create_mock_bucket():
    """Returns a mock bucket

    Args:

    Returns:

    """
    mock_bucket = Mock(spec=Bucket)
    mock_bucket.label = 'bucket'
    mock_bucket.color = '#000000'
    mock_bucket.types = []
    mock_bucket.id = ObjectId()
    return mock_bucket


def _create_bucket():
    """Returns a bucket

    Args:

    Returns:

    """
    return Bucket(
        id=ObjectId(),
        label="bucket",
        color="#000000",
        types=[]
    )


def _create_mock_type_version_manager(title="", versions=None, user='1'):
    """Returns a mock type version manager

    Args:
        title:
        versions:

    Returns:

    """
    mock_type_version_manager = Mock(spec=TypeVersionManager)
    mock_type_version_manager.title = title
    mock_type_version_manager.id = ObjectId()
    mock_type_version_manager.user = user
    if versions is not None:
        mock_type_version_manager.versions = versions
    else:
        mock_type_version_manager.versions = []
    mock_type_version_manager.disabled_versions = []
    return mock_type_version_manager



""" Initialize permissions for core composer app
"""
from django.contrib.auth.models import Permission, Group
from core_main_app.permissions import rights as main_rights
from core_composer_app.permissions import rights as composer_rights


def init_permissions():
    """Initialization of groups and permissions.

    Returns:

    """
    try:
        # Get or Create the default group
        default_group, created = Group.objects.get_or_create(name=main_rights.default_group)

        # Get composer permissions
        composer_access_perm = Permission.objects.get(codename=composer_rights.composer_access)
        composer_save_template_perm = Permission.objects.get(codename=composer_rights.composer_save_template)
        composer_save_type_perm = Permission.objects.get(codename=composer_rights.composer_save_type)

        # add permissions to default group
        default_group.permissions.add(composer_access_perm,
                                      composer_save_template_perm,
                                      composer_save_type_perm)
    except Exception, e:
        print('ERROR : Impossible to init the rules : ' + e.message)



"""Url router for the REST API
"""
from django.conf.urls import url

from core_composer_app.rest.bucket import views as bucket_views
from core_composer_app.rest.type_version_manager import views as type_version_manager_views
from core_main_app.rest.template import views as template_views
from core_main_app.rest.template_version_manager import views as template_version_manager_views

urlpatterns = [
    url(r'^type-version-manager/global/$',
        type_version_manager_views.GlobalTypeVersionManagerList.as_view(),
        name='core_composer_app_rest_type_version_manager_global_list'),

    url(r'^type-version-manager/user/$',
        type_version_manager_views.UserTypeVersionManagerList.as_view(),
        name='core_composer_app_rest_type_version_manager_user_list'),

    url(r'^type-version-manager/(?P<pk>\w+)/$',
        template_version_manager_views.TemplateVersionManagerDetail.as_view(),
        name='core_composer_app_rest_type_version_manager_detail'),

    url(r'^type-version-manager/(?P<pk>\w+)/disable/$',
        template_version_manager_views.DisableTemplateVersionManager.as_view(),
        name='core_composer_app_rest_type_version_manager_disable'),

    url(r'^type-version-manager/(?P<pk>\w+)/restore/$',
        template_version_manager_views.RestoreTemplateVersionManager.as_view(),
        name='core_composer_app_rest_type_version_manager_restore'),

    url(r'^type/version/(?P<pk>\w+)/current/$',
        template_version_manager_views.CurrentTemplateVersion.as_view(),
        name='core_composer_app_rest_type_version_current'),

    url(r'^type/version/(?P<pk>\w+)/disable/$',
        template_version_manager_views.DisableTemplateVersion.as_view(),
        name='core_composer_app_rest_type_version_disable'),

    url(r'^type/global/$',
        type_version_manager_views.GlobalTypeList.as_view(),
        name='core_composer_app_rest_global_type_list'),

    url(r'^type/$',
        type_version_manager_views.UserTypeList.as_view(),
        name='core_composer_app_rest_user_type_list'),

    url(r'^type/version/(?P<pk>\w+)/restore/$',
        template_version_manager_views.RestoreTemplateVersion.as_view(),
        name='core_composer_app_rest_type_version_restore'),

    url(r'^type/(?P<pk>\w+)/download/$', template_views.TemplateDownload.as_view(),
        name='core_composer_app_rest_type_download'),

    url(r'^type/(?P<pk>\w+)/$', template_views.TemplateDetail.as_view(),
        name='core_composer_app_rest_type_detail'),

    url(r'^bucket/$',
        bucket_views.BucketList.as_view(),
        name='core_composer_app_rest_bucket_list'),

    url(r'^bucket/(?P<pk>\w+)/$',
        bucket_views.BucketDetail.as_view(),
        name='core_composer_app_rest_bucket_detail'),

    url(r'^buckets/type-version-manager/(?P<pk>\w+)/$',
        bucket_views.TypeVersionManagerBuckets.as_view(),
        name='core_composer_app_rest_type_version_manger_buckets'),
]



""" Fixtures files for type version manager
"""
from core_composer_app.components.bucket.models import Bucket
from tests.components.type_version_manager.fixtures.fixtures import TypeVersionManagerFixtures


class BucketFixtures(TypeVersionManagerFixtures):
    """ Bucket fixtures
    """
    bucket_empty = None
    bucket_1 = None
    bucket_2 = None
    bucket_collection = None

    def insert_data(self):
        """ Insert a set of Buckets.

        Returns:

        """
        super(BucketFixtures, self).insert_data()

        self.bucket_empty = Bucket(label="empty",
                                   color='#000000',
                                   types=[]).save()
        self.bucket_1 = Bucket(label='bucket1',
                               color='#000001',
                               types=[self.type_vm_1]).save()
        self.bucket_2 = Bucket(label='bucket2',
                               color='#000002',
                               types=[self.type_vm_1, self.type_vm_2]).save()

        self.bucket_collection = [self.bucket_empty,
                                  self.bucket_1,
                                  self.bucket_2]



"""
Type models
"""
from core_main_app.commons import exceptions
from core_main_app.components.template.models import Template
from django_mongoengine import fields
from mongoengine import errors as mongoengine_errors


class Type(Template):
    """Type class.
    """
    is_complex = fields.BooleanField(blank=False)

    @staticmethod
    def get_by_id(type_id):
        """Return a type given its id.

        Args:
            type_id:

        Returns:

        """
        try:
            return Type.objects().get(pk=str(type_id))
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(e.message)
        except Exception as e:
            raise exceptions.ModelError(e.message)

    @staticmethod
    def get_all():
        """Return all types.

        Returns:

        """
        return Type.objects().all()

    @staticmethod
    def get_all_complex_type():
        """List all complex types.

        Returns:

        """
        return Type.objects(is_complex=True).all()



"""Unit tests for composer xml operations
"""
# TODO: add tests for other xml utils

from unittest.case import TestCase
from os.path import join, dirname, abspath
from core_composer_app.utils.xml import _insert_element_type, check_type_core_support, \
    COMPLEX_TYPE, SIMPLE_TYPE
from core_main_app.commons.exceptions import CoreError
from core_main_app.utils.xml import validate_xml_schema

RESOURCES_PATH = join(dirname(abspath(__file__)), 'data')


class TestInsertElementType(TestCase):
    def setUp(self):
        self.root_xpath = 'xsd:complexType/xsd:sequence'
        self.type_name = 'new'

    def test_add_type_to_base(self):
        base_filename = 'base.xsd'
        type_filename = 'type.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(base_content,
                                           self.root_xpath,
                                           type_content,
                                           self.type_name,
                                           join(RESOURCES_PATH, type_filename))

        errors = validate_xml_schema(result_tree)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_to_base(self):
        base_filename = 'base.xsd'
        type_filename = 'type_target_ns.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        with self.assertRaises(CoreError):
            _insert_element_type(base_content,
                                 self.root_xpath,
                                 type_content,
                                 self.type_name,
                                 join(RESOURCES_PATH, type_filename))

    def test_add_type_to_base_with_target_namespace(self):
        base_filename = 'base_target_ns.xsd'
        type_filename = 'type.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(base_content,
                                           self.root_xpath,
                                           type_content,
                                           self.type_name,
                                           join(RESOURCES_PATH, type_filename))

        errors = validate_xml_schema(result_tree)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_to_base_with_same_target_namespace(self):
        base_filename = 'base_target_ns.xsd'
        type_filename = 'type_base_target_ns.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(base_content,
                                           self.root_xpath,
                                           type_content,
                                           self.type_name,
                                           join(RESOURCES_PATH, type_filename))

        errors = validate_xml_schema(result_tree)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_to_base_with_different_target_namespace(self):
        base_filename = 'base_target_ns.xsd'
        type_filename = 'type_target_ns.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        with self.assertRaises(CoreError):
            _insert_element_type(base_content,
                                 self.root_xpath,
                                 type_content,
                                 self.type_name,
                                 join(RESOURCES_PATH, type_filename))

    def test_add_type_with_target_namespace_prefix_to_base(self):
        base_filename = 'base.xsd'
        type_filename = 'type_target_ns_prefix.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(base_content,
                                           self.root_xpath,
                                           type_content,
                                           self.type_name,
                                           join(RESOURCES_PATH, type_filename))

        errors = validate_xml_schema(result_tree)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_prefix_to_base_with_same_target_namespace_prefix(self):
        base_filename = 'base_target_ns.xsd'
        type_filename = 'type_base_target_ns_prefix.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(base_content,
                                           self.root_xpath,
                                           type_content,
                                           self.type_name,
                                           join(RESOURCES_PATH, type_filename))

        errors = validate_xml_schema(result_tree)
        self.assertTrue(errors is None)

    def test_add_type_with_target_namespace_prefix_to_base_with_different_target_namespace_prefix(self):
        base_filename = 'base_target_ns.xsd'
        type_filename = 'type_target_ns_prefix.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, base_filename), 'r') as base_file:
            base_content = base_file.read()
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        result_tree = _insert_element_type(base_content,
                                           self.root_xpath,
                                           type_content,
                                           self.type_name,
                                           join(RESOURCES_PATH, type_filename))

        errors = validate_xml_schema(result_tree)
        self.assertTrue(errors is None)


class TestTypeDefinition(TestCase):
    def test_simple_type(self):
        type_filename = 'type.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        type_content = check_type_core_support(type_content)

        self.assertEqual(type_content, SIMPLE_TYPE)

    def test_complex_type(self):
        type_filename = 'type_complex.xsd'
        # load test resources
        with open(join(RESOURCES_PATH, type_filename), 'r') as type_file:
            type_content = type_file.read()

        type_content = check_type_core_support(type_content)

        self.assertEqual(type_content, COMPLEX_TYPE)


"""
Type Version Manager API
"""
from core_composer_app.components.type import api as type_api
from core_composer_app.components.type_version_manager.models import TypeVersionManager
from core_composer_app.components.bucket import api as bucket_api
from core_main_app.components.template import api as template_api
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.components.version_manager.utils import get_latest_version_name


def insert(type_version_manager, type_object, list_bucket_ids=None):
    """Add a version to a type version manager.

    Args:
        type_version_manager:
        type_object:
        list_bucket_ids:

    Returns:

    """
    # save the type in database
    type_api.upsert(type_object)
    try:
        # insert the initial template in the version manager
        version_manager_api.insert_version(type_version_manager, type_object)
        # insert the version manager in database
        version_manager_api.upsert(type_version_manager)
        # Add to the selected buckets
        bucket_api.add_type_to_buckets(type_version_manager, list_bucket_ids)
        # get type display name
        display_name = get_latest_version_name(type_version_manager)
        # update saved template
        template_api.set_display_name(type_object, display_name)
        # update saved type
        return type_version_manager
    except Exception, e:
        type_api.delete(type_object)
        raise e


def get_global_version_managers():
    """Get all global version managers of a type.

    Returns:

    """
    return TypeVersionManager.get_global_version_managers()


def get_active_global_version_manager():
    """ Return all active Version Managers with user set to None.

    Returns:

    """
    return TypeVersionManager.get_active_global_version_manager()


def get_version_managers_by_user(user_id):
    """Get all global version managers of a user.

    Returns:

    """
    return TypeVersionManager.get_version_managers_by_user(user_id)


def get_active_version_manager_by_user_id(user_id):
    """ Return all active Version Managers with given user id.

    Returns:

    """
    return TypeVersionManager.get_active_version_manager_by_user_id(user_id)


def get_no_buckets_types():
    """Get list of available types not inside a bucket.

    Returns:

    """
    # build list of types
    bucket_types = []
    for bucket in bucket_api.get_all():
        bucket_types += bucket.types

    all_types = get_global_version_managers()
    no_bucket_types = [type_version_manager for type_version_manager in all_types
                       if type_version_manager not in bucket_types]
    return no_bucket_types


def get_all_version_manager_except_user_id(user_id):
    """ Return all Type Version Managers of all users except user with given user id.

    Returns:

    """
    return TypeVersionManager.get_all_type_version_manager_except_user_id(user_id)


def get_all_version_manager():
    """ Return all Type Version Managers of all users.

    Returns:

    """
    return TypeVersionManager.get_all_type_version_manager()



"""Composer Admin Forms
"""
from django import forms
from mongodbforms import DocumentForm

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.components.bucket.models import Bucket


class BucketForm(forms.Form):
    """
    Form to add a bucket.
    """
    label = forms.CharField(label='Enter Bucket label', max_length=100, required=True,
                            widget=forms.TextInput(attrs={'class': 'form-control'}))


class BucketDataModelChoiceField(forms.ModelMultipleChoiceField):
    """
    Choice Field to select an existing form.
    """
    def label_from_instance(self, obj):
        return obj.label


class UploadTypeForm(forms.Form):
    """
    Form to upload a new Type.
    """
    name = forms.CharField(label='Enter Type name', max_length=100, required=True,
                           widget=forms.TextInput(attrs={'class': 'form-control'}))
    xsd_file = forms.FileField(label='Select a file', required=True,
                               widget=forms.FileInput(attrs={'class': 'form-control'}))
    buckets = BucketDataModelChoiceField(label='Select buckets', queryset=bucket_api.get_all(), required=False,
                                         widget=forms.SelectMultiple(attrs={'class': 'form-control'}))


class EditTypeBucketsForm(forms.Form):
    """
    Form to edit buckets of a Type.
    """
    buckets = BucketDataModelChoiceField(label='Select new buckets', queryset=bucket_api.get_all(), required=False,
                                         widget=forms.SelectMultiple(attrs={'class': 'form-control'}))


class EditBucketForm(DocumentForm):
    label = forms.CharField(label='Label',
                            widget=forms.TextInput(attrs={'class': 'form-control',
                                                          'placeholder': 'Enter the new label'}))

    class Meta:
        document = Bucket
        fields = ['label']



""" Views for the Bucket REST API
"""
from django.http import Http404
from django.utils.decorators import method_decorator
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView

from core_composer_app.components.bucket import api as bucket_api
from core_composer_app.rest.bucket.serializers import BucketSerializer, BucketsSerializer
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.rest.template_version_manager.abstract_views import AbstractTemplateVersionManagerDetail
from core_main_app.utils.access_control.exceptions import AccessControlError
from core_main_app.utils.decorators import api_staff_member_required


class BucketList(APIView):

    def get(self, request):
        """ Get buckets.

        /rest/bucket/
        /rest/bucket/?label=<label>

        Query Params:
            label: label

        Args:
            request:

        Returns:

        """
        try:
            # Get objects
            object_list = bucket_api.get_all()

            # Apply filters
            label = self.request.query_params.get('label', None)
            if label is not None:
                object_list = object_list.filter(label=label)

            # Serialize object
            serializer = BucketSerializer(object_list, many=True)
            return Response(serializer.data, status=status.HTTP_200_OK)
        except Exception as api_exception:
            content = {'message': api_exception.message}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @method_decorator(api_staff_member_required())
    def post(self, request):
        """ Create bucket

        /rest/bucket/

        Data:
            {
            "label": "label",
            "types": ["<type_version_manager_id>"]
            }

        Args:
            request:

        Returns:

        """
        try:
            # Build serializer
            bucket_serializer = BucketSerializer(data=request.data)

            # Validate data
            bucket_serializer.is_valid(True)
            # Save data
            bucket_serializer.save()

            # Return the serialized data
            return Response(bucket_serializer.data, status=status.HTTP_201_CREATED)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': api_exception.message}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class BucketDetail(APIView):
    """
    Retrieve, update or delete a bucket.
    """

    def get_object(self, pk):
        """ Get data from db

        Args:
            pk:

        Returns:

        """
        try:
            return bucket_api.get_by_id(pk)
        except DoesNotExist:
            raise Http404

    def get(self, request, pk):
        """ Retrieve bucket

        Args:
            request:
            pk:

        Returns:

        """
        try:
            # Get object
            bucket = self.get_object(pk)

            # Serialize object
            serializer = BucketSerializer(bucket)

            # Return response
            return Response(serializer.data)
        except Http404:
            content = {'message': 'Bucket not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': api_exception.message}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @method_decorator(api_staff_member_required())
    def delete(self, request, pk):
        """ Delete a bucket

        Args:
            request:
            pk:

        Returns:

        """
        try:
            # Get object
            bucket = self.get_object(pk)

            # delete object
            bucket_api.delete(bucket)

            # Return response
            return Response(status=status.HTTP_204_NO_CONTENT)
        except Http404:
            content = {'message': 'Bucket not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': api_exception.message}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    @method_decorator(api_staff_member_required())
    def patch(self, request, pk):
        """ Update bucket

        Data:
            {"label": "<label>"}

        Args:
            request:
            pk:

        Returns:

        """
        try:
            # Get object
            bucket = self.get_object(pk)

            # Build serializer
            bucket_serializer = BucketSerializer(instance=bucket,
                                                 data=request.data,
                                                 partial=True)

            # Validate data
            bucket_serializer.is_valid(True)
            # Save data
            bucket_serializer.save(user=request.user)

            return Response(bucket_serializer.data, status=status.HTTP_200_OK)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Http404:
            content = {'message': 'Bucket not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except Exception as api_exception:
            content = {'message': api_exception.message}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


class TypeVersionManagerBuckets(AbstractTemplateVersionManagerDetail):
    @method_decorator(api_staff_member_required())
    def patch(self, request, pk):
        """ Set new list of buckets for a type version manager.

        /rest/buckets/type-version-manager/<tvm_id>/

        Data:
        [{"id":"<bucket_id>"},{"id":"<bucket_id>"}]

        Args:
            request:
            pk:

        Returns:

        """
        try:
            # Get object
            type_version_manager = self.get_object(pk)

            # Serialize data
            serializer = BucketsSerializer(data=request.data, many=True)

            # Validate data
            serializer.is_valid(True)

            # Get list of unique ids
            bucket_ids = set([bucket['id'] for bucket in serializer.validated_data])

            # act
            bucket_api.update_type_buckets(type_version_manager, bucket_ids)

            return Response(status=status.HTTP_200_OK)
        except Http404:
            content = {'message': 'Object not found.'}
            return Response(content, status=status.HTTP_404_NOT_FOUND)
        except AccessControlError as access_error:
            content = {'message': access_error.message}
            return Response(content, status=status.HTTP_403_FORBIDDEN)
        except ValidationError as validation_exception:
            content = {'message': validation_exception.detail}
            return Response(content, status=status.HTTP_400_BAD_REQUEST)
        except Exception as api_exception:
            content = {'message': api_exception.message}
            return Response(content, status=status.HTTP_500_INTERNAL_SERVER_ERROR)

